import numpy as np

# conversions
J_to_atomic = 1.0 / 4.359744e-18
atomic_to_kJmol = 2625.499639
atomic_to_K = 315775.04e0
BOHR_to_ANGSTROM = 0.52917721092e0
DEGREES_per_RADIAN = 180.0 / np.pi
atomic_to_kg = 1.66054e-27
atomic_to_fs = 1.0327503e0

# constants
kB_in_SI = 1.380648e-23
R_in_SI = 8.314
kB_in_atomic = kB_in_SI * J_to_atomic
h_in_SI = 6.62607004e-34
